# copyright 2003-2011 LOGILAB S.A. (Paris, FRANCE), all rights reserved.
# contact http://www.logilab.fr/ -- mailto:contact@logilab.fr
#
# This file is part of logilab-common.
#
# logilab-common is free software: you can redistribute it and/or modify it under
# the terms of the GNU Lesser General Public License as published by the Free
# Software Foundation, either version 2.1 of the License, or (at your option) any
# later version.
#
# logilab-common is distributed in the hope that it will be useful, but WITHOUT
# ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
# FOR A PARTICULAR PURPOSE.  See the GNU Lesser General Public License for more
# details.
#
# You should have received a copy of the GNU Lesser General Public License along
# with logilab-common.  If not, see <http://www.gnu.org/licenses/>.
"""Some text manipulation utility functions.


:group text formatting: normalize_text, normalize_paragraph, pretty_match,\
unquote, colorize_ansi
:group text manipulation: searchall, splitstrip
:sort: text formatting, text manipulation

:type ANSI_STYLES: dict(str)
:var ANSI_STYLES: dictionary mapping style identifier to ANSI terminal code

:type ANSI_COLORS: dict(str)
:var ANSI_COLORS: dictionary mapping color identifier to ANSI terminal code

:type ANSI_PREFIX: str
:var ANSI_PREFIX:
  ANSI terminal code notifying the start of an ANSI escape sequence

:type ANSI_END: str
:var ANSI_END:
  ANSI terminal code notifying the end of an ANSI escape sequence

:type ANSI_RESET: str
:var ANSI_RESET:
  ANSI terminal code resetting format defined by a previous ANSI escape sequence
"""
__docformat__ = "restructuredtext en"

import sys
import re
import os.path as osp

try:
    from re import Pattern, Match  # type: ignore
except ImportError:
    # Pattern and Match are python > 3.6 only.
    #
    # To be compatible with python <= 3.6, and still provide some typing, we
    # manually define Pattern and Match, in the same manner they are defined in
    # the re module of python > 3.7
    # cf https://github.com/python/cpython/blob/3.7/Lib/re.py#L264
    Pattern = type(re.sre_compile.compile("", 0))  # type: ignore
    Match = type(re.sre_compile.compile("", 0).match(""))  # type: ignore
from warnings import warn
from unicodedata import normalize as _uninormalize
from typing import Any, Optional, Tuple, List, Callable, Dict, Union

try:
    from os import linesep
except ImportError:
    linesep = "\n"  # gae

from logilab.common.deprecation import callable_deprecated

MANUAL_UNICODE_MAP = {
    "\xa1": "!",  # INVERTED EXCLAMATION MARK
    "\u0142": "l",  # LATIN SMALL LETTER L WITH STROKE
    "\u2044": "/",  # FRACTION SLASH
    "\xc6": "AE",  # LATIN CAPITAL LETTER AE
    "\xa9": "(c)",  # COPYRIGHT SIGN
    "\xab": '"',  # LEFT-POINTING DOUBLE ANGLE QUOTATION MARK
    "\xe6": "ae",  # LATIN SMALL LETTER AE
    "\xae": "(r)",  # REGISTERED SIGN
    "\u0153": "oe",  # LATIN SMALL LIGATURE OE
    "\u0152": "OE",  # LATIN CAPITAL LIGATURE OE
    "\xd8": "O",  # LATIN CAPITAL LETTER O WITH STROKE
    "\xf8": "o",  # LATIN SMALL LETTER O WITH STROKE
    "\xbb": '"',  # RIGHT-POINTING DOUBLE ANGLE QUOTATION MARK
    "\xdf": "ss",  # LATIN SMALL LETTER SHARP S
    "\u2013": "-",  # HYPHEN
    "\u2019": "'",  # SIMPLE QUOTE
}


def unormalize(
    ustring: str, ignorenonascii: Optional[Any] = None, substitute: Optional[str] = None
) -> str:
    """replace diacritical characters with their corresponding ascii characters

    Convert the unicode string to its long normalized form (unicode character
    will be transform into several characters) and keep the first one only.
    The normal form KD (NFKD) will apply the compatibility decomposition, i.e.
    replace all compatibility characters with their equivalents.

    :type substitute: str
    :param substitute: replacement character to use if decomposition fails

    :see: Another project about ASCII transliterations of Unicode text
          http://pypi.python.org/pypi/Unidecode
    """
    # backward compatibility, ignorenonascii was a boolean
    if ignorenonascii is not None:
        warn(
            "ignorenonascii is deprecated, use substitute named parameter instead",
            DeprecationWarning,
            stacklevel=2,
        )
        if ignorenonascii:
            substitute = ""
    res = []
    for letter in ustring[:]:
        try:
            replacement = MANUAL_UNICODE_MAP[letter]
        except KeyError:
            replacement = _uninormalize("NFKD", letter)[0]
            if ord(replacement) >= 2 ** 7:
                if substitute is None:
                    raise ValueError("can't deal with non-ascii based characters")
                replacement = substitute
        res.append(replacement)
    return "".join(res)


def unquote(string: str) -> str:
    """remove optional quotes (simple or double) from the string

    :type string: str or unicode
    :param string: an optionally quoted string

    :rtype: str or unicode
    :return: the unquoted string (or the input string if it wasn't quoted)
    """
    if not string:
        return string
    if string[0] in "\"'":
        string = string[1:]
    if string[-1] in "\"'":
        string = string[:-1]
    return string


_BLANKLINES_RGX = re.compile("\r?\n\r?\n")
_NORM_SPACES_RGX = re.compile(r"\s+")


def normalize_text(text: str, line_len: int = 80, indent: str = "", rest: bool = False) -> str:
    """normalize a text to display it with a maximum line size and
    optionally arbitrary indentation. Line jumps are normalized but blank
    lines are kept. The indentation string may be used to insert a
    comment (#) or a quoting (>) mark  for instance.

    :type text: str or unicode
    :param text: the input text to normalize

    :type line_len: int
    :param line_len: expected maximum line's length, default to 80

    :type indent: str or unicode
    :param indent: optional string to use as indentation

    :rtype: str or unicode
    :return:
      the input text normalized to fit on lines with a maximized size
      inferior to `line_len`, and optionally prefixed by an
      indentation string
    """
    if rest:
        normp = normalize_rest_paragraph
    else:
        normp = normalize_paragraph
    result = []
    for text in _BLANKLINES_RGX.split(text):
        result.append(normp(text, line_len, indent))
    return ("%s%s%s" % (linesep, indent, linesep)).join(result)


def normalize_paragraph(text: str, line_len: int = 80, indent: str = "") -> str:
    """normalize a text to display it with a maximum line size and
    optionally arbitrary indentation. Line jumps are normalized. The
    indentation string may be used top insert a comment mark for
    instance.

    :type text: str or unicode
    :param text: the input text to normalize

    :type line_len: int
    :param line_len: expected maximum line's length, default to 80

    :type indent: str or unicode
    :param indent: optional string to use as indentation

    :rtype: str or unicode
    :return:
      the input text normalized to fit on lines with a maximized size
      inferior to `line_len`, and optionally prefixed by an
      indentation string
    """
    text = _NORM_SPACES_RGX.sub(" ", text)
    line_len = line_len - len(indent)
    lines = []
    while text:
        aline, text = splittext(text.strip(), line_len)
        lines.append(indent + aline)
    return linesep.join(lines)


def normalize_rest_paragraph(text: str, line_len: int = 80, indent: str = "") -> str:
    """normalize a ReST text to display it with a maximum line size and
    optionally arbitrary indentation. Line jumps are normalized. The
    indentation string may be used top insert a comment mark for
    instance.

    :type text: str or unicode
    :param text: the input text to normalize

    :type line_len: int
    :param line_len: expected maximum line's length, default to 80

    :type indent: str or unicode
    :param indent: optional string to use as indentation

    :rtype: str or unicode
    :return:
      the input text normalized to fit on lines with a maximized size
      inferior to `line_len`, and optionally prefixed by an
      indentation string
    """
    toreport = ""
    lines = []
    line_len = line_len - len(indent)
    for line in text.splitlines():
        line = toreport + _NORM_SPACES_RGX.sub(" ", line.strip())
        toreport = ""
        while len(line) > line_len:
            # too long line, need split
            line, toreport = splittext(line, line_len)
            lines.append(indent + line)
            if toreport:
                line = toreport + " "
                toreport = ""
            else:
                line = ""
        if line:
            lines.append(indent + line.strip())
    return linesep.join(lines)


def splittext(text: str, line_len: int) -> Tuple[str, str]:
    """split the given text on space according to the given max line size

    return a 2-uple:
    * a line <= line_len if possible
    * the rest of the text which has to be reported on another line
    """
    if len(text) <= line_len:
        return text, ""
    pos = min(len(text) - 1, line_len)
    while pos > 0 and text[pos] != " ":
        pos -= 1
    if pos == 0:
        pos = min(len(text), line_len)
        while len(text) > pos and text[pos] != " ":
            pos += 1
    return text[:pos], text[pos + 1 :].strip()


def splitstrip(string: str, sep: str = ",") -> List[str]:
    """return a list of stripped string by splitting the string given as
    argument on `sep` (',' by default). Empty string are discarded.

    >>> splitstrip('a, b, c   ,  4,,')
    ['a', 'b', 'c', '4']
    >>> splitstrip('a')
    ['a']
    >>>

    :type string: str or unicode
    :param string: a csv line

    :type sep: str or unicode
    :param sep: field separator, default to the comma (',')

    :rtype: str or unicode
    :return: the unquoted string (or the input string if it wasn't quoted)
    """
    return [word.strip() for word in string.split(sep) if word.strip()]


get_csv = callable_deprecated("get_csv is deprecated, use splitstrip")(splitstrip)


def split_url_or_path(url_or_path):
    """return the latest component of a string containing either an url of the
    form <scheme>://<path> or a local file system path
    """
    if "://" in url_or_path:
        return url_or_path.rstrip("/").rsplit("/", 1)
    return osp.split(url_or_path.rstrip(osp.sep))


def text_to_dict(text):
    """parse multilines text containing simple 'key=value' lines and return a
    dict of {'key': 'value'}. When the same key is encountered multiple time,
    value is turned into a list containing all values.

    >>> d = text_to_dict('''multiple=1
    ... multiple= 2
    ... single =3
    ... ''')
    >>> d['single']
    '3'
    >>> d['multiple']
    ['1', '2']

    """
    res = {}
    if not text:
        return res
    for line in text.splitlines():
        line = line.strip()
        if line and not line.startswith("#"):
            key, value = [w.strip() for w in line.split("=", 1)]
            if key in res:
                try:
                    res[key].append(value)
                except AttributeError:
                    res[key] = [res[key], value]
            else:
                res[key] = value
    return res


_BLANK_URE = r"(\s|,)+"
_BLANK_RE = re.compile(_BLANK_URE)
__VALUE_URE = r"-?(([0-9]+\.[0-9]*)|((0x?)?[0-9]+))"
__UNITS_URE = r"[a-zA-Z]+"
_VALUE_RE = re.compile(r"(?P<value>%s)(?P<unit>%s)?" % (__VALUE_URE, __UNITS_URE))
_VALIDATION_RE = re.compile(r"^((%s)(%s))*(%s)?$" % (__VALUE_URE, __UNITS_URE, __VALUE_URE))

BYTE_UNITS = {
    "b": 1,
    "kb": 1024,
    "mb": 1024 ** 2,
    "gb": 1024 ** 3,
    "tb": 1024 ** 4,
}

TIME_UNITS = {
    "ms": 0.0001,
    "s": 1,
    "min": 60,
    "h": 60 * 60,
    "d": 60 * 60 * 24,
}


def apply_units(
    string: str,
    units: Dict[str, int],
    inter: Union[Callable, None, type] = None,
    final: type = float,
    blank_reg: Pattern = _BLANK_RE,
    value_reg: Pattern = _VALUE_RE,
) -> Union[float, int]:
    """Parse the string applying the units defined in units
    (e.g.: "1.5m",{'m',60} -> 80).

    :type string: str or unicode
    :param string: the string to parse

    :type units: dict (or any object with __getitem__ using basestring key)
    :param units: a dict mapping a unit string repr to its value

    :type inter: type
    :param inter: used to parse every intermediate value (need __sum__)

    :type blank_reg: regexp
    :param blank_reg: should match every blank char to ignore.

    :type value_reg: regexp with "value" and optional "unit" group
    :param value_reg: match a value and it's unit into the
    """
    if inter is None:
        inter = final
    fstring = _BLANK_RE.sub("", string)
    if not (fstring and _VALIDATION_RE.match(fstring)):
        raise ValueError("Invalid unit string: %r." % string)
    values = []
    for match in value_reg.finditer(fstring):
        dic = match.groupdict()
        lit, unit = dic["value"], dic.get("unit")
        value = inter(lit)
        if unit is not None:
            try:
                value *= units[unit.lower()]
            except KeyError:
                raise ValueError("invalid unit %s. valid units are %s" % (unit, list(units.keys())))
        values.append(value)
    return final(sum(values))


_LINE_RGX = re.compile("\r\n|\r+|\n")


def pretty_match(match: Match, string: str, underline_char: str = "^") -> str:
    """return a string with the match location underlined:

    >>> import re
    >>> print(pretty_match(re.search('mange', 'il mange du bacon'), 'il mange du bacon'))
    il mange du bacon
       ^^^^^
    >>>

    :type match: _sre.SRE_match
    :param match: object returned by re.match, re.search or re.finditer

    :type string: str or unicode
    :param string:
      the string on which the regular expression has been applied to
      obtain the `match` object

    :type underline_char: str or unicode
    :param underline_char:
      character to use to underline the matched section, default to the
      carret '^'

    :rtype: str or unicode
    :return:
      the original string with an inserted line to underline the match
      location
    """
    start = match.start()
    end = match.end()
    string = _LINE_RGX.sub(linesep, string)
    start_line_pos = string.rfind(linesep, 0, start)
    if start_line_pos == -1:
        start_line_pos = 0
        result = []
    else:
        result = [string[:start_line_pos]]
        start_line_pos += len(linesep)
    offset = start - start_line_pos
    underline = " " * offset + underline_char * (end - start)
    end_line_pos = string.find(linesep, end)
    if end_line_pos == -1:
        string = string[start_line_pos:]
        result.append(string)
        result.append(underline)
    else:
        # mypy: Incompatible types in assignment (expression has type "str",
        # mypy: variable has type "int")
        # but it's a str :|
        end = string[end_line_pos + len(linesep) :]  # type: ignore
        string = string[start_line_pos:end_line_pos]
        result.append(string)
        result.append(underline)
        result.append(end)  # type: ignore # see previous comment
    return linesep.join(result).rstrip()


# Ansi colorization ###########################################################

ANSI_PREFIX = "\033["
ANSI_END = "m"
ANSI_RESET = "\033[0m"
ANSI_STYLES = {
    "reset": "0",
    "bold": "1",
    "italic": "3",
    "underline": "4",
    "blink": "5",
    "inverse": "7",
    "strike": "9",
}
ANSI_COLORS = {
    "reset": "0",
    "black": "30",
    "red": "31",
    "green": "32",
    "yellow": "33",
    "blue": "34",
    "magenta": "35",
    "cyan": "36",
    "white": "37",
}


def _get_ansi_code(color: Optional[str] = None, style: Optional[str] = None) -> str:
    """return ansi escape code corresponding to color and style

    :type color: str or None
    :param color:
      the color name (see `ANSI_COLORS` for available values)
      or the color number when 256 colors are available

    :type style: str or None
    :param style:
      style string (see `ANSI_COLORS` for available values). To get
      several style effects at the same time, use a coma as separator.

    :raise KeyError: if an unexistent color or style identifier is given

    :rtype: str
    :return: the built escape code
    """
    ansi_code = []
    if style:
        style_attrs = splitstrip(style)
        for effect in style_attrs:
            ansi_code.append(ANSI_STYLES[effect])
    if color:
        if color.isdigit():
            ansi_code.extend(["38", "5"])
            ansi_code.append(color)
        else:
            ansi_code.append(ANSI_COLORS[color])
    if ansi_code:
        return ANSI_PREFIX + ";".join(ansi_code) + ANSI_END
    return ""


def colorize_ansi(msg: str, color: Optional[str] = None, style: Optional[str] = None) -> str:
    """colorize message by wrapping it with ansi escape codes

    :type msg: str or unicode
    :param msg: the message string to colorize

    :type color: str or None
    :param color:
      the color identifier (see `ANSI_COLORS` for available values)

    :type style: str or None
    :param style:
      style string (see `ANSI_COLORS` for available values). To get
      several style effects at the same time, use a coma as separator.

    :raise KeyError: if an unexistent color or style identifier is given

    :rtype: str or unicode
    :return: the ansi escaped string
    """
    # If both color and style are not defined, then leave the text as is
    if color is None and style is None:
        return msg
    escape_code = _get_ansi_code(color, style)
    # If invalid (or unknown) color, don't wrap msg with ansi codes
    if escape_code:
        return "%s%s%s" % (escape_code, msg, ANSI_RESET)
    return msg


DIFF_STYLE = {"separator": "cyan", "remove": "red", "add": "green"}


def diff_colorize_ansi(lines, out=sys.stdout, style=DIFF_STYLE):
    for line in lines:
        if line[:4] in ("--- ", "+++ "):
            out.write(colorize_ansi(line, style["separator"]))
        elif line[0] == "-":
            out.write(colorize_ansi(line, style["remove"]))
        elif line[0] == "+":
            out.write(colorize_ansi(line, style["add"]))
        elif line[:4] == "--- ":
            out.write(colorize_ansi(line, style["separator"]))
        elif line[:4] == "+++ ":
            out.write(colorize_ansi(line, style["separator"]))
        else:
            out.write(line)



# copyright 2003-2011 LOGILAB S.A. (Paris, FRANCE), all rights reserved.
# contact http://www.logilab.fr/ -- mailto:contact@logilab.fr
#
# This file is part of logilab-common.
#
# logilab-common is free software: you can redistribute it and/or modify it
# under the terms of the GNU Lesser General Public License as published by the
# Free Software Foundation, either version 2.1 of the License,
# or (at your option) any later version.
#
# logilab-common is distributed in the hope that it will be useful, but WITHOUT
# ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
# FOR A PARTICULAR PURPOSE.  See the GNU Lesser General Public License for more
# details.
#
# You should have received a copy of the GNU Lesser General Public License
# along with logilab-common.  If not, see <http://www.gnu.org/licenses/>.
"""Universal report objects and some formatting drivers.

A way to create simple reports using python objects, primarily designed to be
formatted as text and html.
"""
__docformat__ = "restructuredtext en"

import sys

from typing import Any, Optional, Union, List as ListType, Generator, Tuple, Callable, TextIO

from logilab.common.compat import StringIO
from logilab.common.textutils import linesep
from logilab.common.tree import VNode
from logilab.common.ureports.nodes import Table, Section, Link, Paragraph, Title, Text

from logilab.common.ureports.nodes import VerbatimText, Image, Span, List  # noqa


def get_nodes(node, klass):
    """return an iterator on all children node of the given klass"""
    for child in node.children:
        if isinstance(child, klass):
            yield child
        # recurse (FIXME: recursion controled by an option)
        for grandchild in get_nodes(child, klass):
            yield grandchild


def layout_title(layout):
    """try to return the layout's title as string, return None if not found"""
    for child in layout.children:
        if isinstance(child, Title):
            return " ".join([node.data for node in get_nodes(child, Text)])


def build_summary(layout, level=1):
    """make a summary for the report, including X level"""
    assert level > 0
    level -= 1
    summary = ListType(klass="summary")
    for child in layout.children:
        if not isinstance(child, Section):
            continue
        label = layout_title(child)
        if not label and not child.id:
            continue
        if not child.id:
            child.id = label.replace(" ", "-")
        node = Link("#" + child.id, label=label or child.id)
        # FIXME: Three following lines produce not very compliant
        # docbook: there are some useless <para><para>. They might be
        # replaced by the three commented lines but this then produces
        # a bug in html display...
        if level and [n for n in child.children if isinstance(n, Section)]:
            node = Paragraph([node, build_summary(child, level)])
        summary.append(node)
    #         summary.append(node)
    #         if level and [n for n in child.children if isinstance(n, Section)]:
    #             summary.append(build_summary(child, level))
    return summary


class BaseWriter(object):
    """base class for ureport writers"""

    def format(
        self,
        layout: Any,
        stream: Optional[Union[StringIO, TextIO]] = None,
        encoding: Optional[Any] = None,
    ) -> None:
        """format and write the given layout into the stream object

        unicode policy: unicode strings may be found in the layout;
        try to call stream.write with it, but give it back encoded using
        the given encoding if it fails
        """
        if stream is None:
            stream = sys.stdout
        if not encoding:
            encoding = getattr(stream, "encoding", "UTF-8")
        self.encoding = encoding or "UTF-8"
        self.__compute_funcs: ListType[Tuple[Callable[[str], Any], Callable[[str], Any]]] = []
        self.out = stream
        self.begin_format(layout)
        layout.accept(self)
        self.end_format(layout)

    def format_children(self, layout: Union["Paragraph", "Section", "Title"]) -> None:
        """recurse on the layout children and call their accept method
        (see the Visitor pattern)
        """
        for child in getattr(layout, "children", ()):
            child.accept(self)

    def writeln(self, string: str = "") -> None:
        """write a line in the output buffer"""
        self.write(string + linesep)

    def write(self, string: str) -> None:
        """write a string in the output buffer"""
        try:
            self.out.write(string)
        except UnicodeEncodeError:
            # mypy: Argument 1 to "write" of "IO" has incompatible type "bytes"; expected "str"
            # probably a python3 port issue?
            self.out.write(string.encode(self.encoding))  # type: ignore

    def begin_format(self, layout: Any) -> None:
        """begin to format a layout"""
        self.section = 0

    def end_format(self, layout: Any) -> None:
        """finished to format a layout"""

    def get_table_content(self, table: Table) -> ListType[ListType[str]]:
        """trick to get table content without actually writing it

        return an aligned list of lists containing table cells values as string
        """
        result: ListType[ListType[str]] = [[]]
        # mypy: "Table" has no attribute "cols"
        # dynamic attribute
        cols = table.cols  # type: ignore

        for cell in self.compute_content(table):
            if cols == 0:
                result.append([])
                # mypy: "Table" has no attribute "cols"
                # dynamic attribute
                cols = table.cols  # type: ignore

            cols -= 1
            result[-1].append(cell)

        # fill missing cells
        while len(result[-1]) < cols:
            result[-1].append("")

        return result

    def compute_content(self, layout: VNode) -> Generator[str, Any, None]:
        """trick to compute the formatting of children layout before actually
        writing it

        return an iterator on strings (one for each child element)
        """
        # use cells !
        def write(data: str) -> None:
            try:
                stream.write(data)
            except UnicodeEncodeError:
                # mypy: Argument 1 to "write" of "TextIOWrapper" has incompatible type "bytes";
                # mypy: expected "str"
                # error from porting to python3?
                stream.write(data.encode(self.encoding))  # type: ignore

        def writeln(data: str = "") -> None:
            try:
                stream.write(data + linesep)
            except UnicodeEncodeError:
                # mypy: Unsupported operand types for + ("bytes" and "str")
                # error from porting to python3?
                stream.write(data.encode(self.encoding) + linesep)  # type: ignore

        # mypy: Cannot assign to a method
        # this really looks like black dirty magic since self.write is reused elsewhere in the code
        # especially since self.write and self.writeln are conditionally
        # deleted at the end of this function
        self.write = write  # type: ignore
        self.writeln = writeln  # type: ignore

        self.__compute_funcs.append((write, writeln))

        # mypy: Item "Table" of "Union[ListType[Any], Table, Title]" has no attribute "children"
        # dynamic attribute?
        for child in layout.children:  # type: ignore
            stream = StringIO()

            child.accept(self)

            yield stream.getvalue()

        self.__compute_funcs.pop()

        try:
            # mypy: Cannot assign to a method
            # even more black dirty magic
            self.write, self.writeln = self.__compute_funcs[-1]  # type: ignore
        except IndexError:
            del self.write
            del self.writeln


from logilab.common.ureports.text_writer import TextWriter  # noqa
from logilab.common.ureports.html_writer import HTMLWriter  # noqa



# copyright 2003-2012 LOGILAB S.A. (Paris, FRANCE), all rights reserved.
# contact http://www.logilab.fr/ -- mailto:contact@logilab.fr
#
# This file is part of logilab-common.
#
# logilab-common is free software: you can redistribute it and/or modify it under
# the terms of the GNU Lesser General Public License as published by the Free
# Software Foundation, either version 2.1 of the License, or (at your option) any
# later version.
#
# logilab-common is distributed in the hope that it will be useful, but WITHOUT
# ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
# FOR A PARTICULAR PURPOSE.  See the GNU Lesser General Public License for more
# details.
#
# You should have received a copy of the GNU Lesser General Public License along
# with logilab-common.  If not, see <http://www.gnu.org/licenses/>.
"""Unicode email support (extends email from stdlib)"""

__docformat__ = "restructuredtext en"

import email
from encodings import search_function
import sys
from email.utils import parseaddr, parsedate
from email.header import decode_header
from email.message import Message

from datetime import datetime
from typing import Any, Optional, List, Tuple, Union

try:
    from mx.DateTime import DateTime
except ImportError:
    DateTime = datetime

import logilab.common as lgc


def decode_QP(string: str) -> str:
    parts: List[str] = []
    for maybe_decoded, charset in decode_header(string):
        if not charset:
            charset = "iso-8859-15"
        # python 3 sometimes returns str and sometimes bytes.
        # the 'official' fix is to use the new 'policy' APIs
        # https://bugs.python.org/issue24797
        # let's just handle this bug ourselves for now
        if isinstance(maybe_decoded, bytes):
            decoded = maybe_decoded.decode(charset, "replace")
        else:
            decoded = maybe_decoded

        assert isinstance(decoded, str)
        parts.append(decoded)

    if sys.version_info < (3, 3):
        # decoding was non-RFC compliant wrt to whitespace handling
        # see http://bugs.python.org/issue1079
        return " ".join(parts)

    return "".join(parts)


def message_from_file(fd):
    try:
        return UMessage(email.message_from_file(fd))
    except email.errors.MessageParseError:
        return ""


def message_from_string(string: str) -> Union["UMessage", str]:
    try:
        return UMessage(email.message_from_string(string))
    except email.errors.MessageParseError:
        return ""


class UMessage:
    """Encapsulates an email.Message instance and returns only unicode objects."""

    def __init__(self, message: Message) -> None:
        self.message = message

    # email.Message interface #################################################

    def get(self, header: str, default: Optional[Any] = None) -> Optional[str]:
        value = self.message.get(header, default)
        if value:
            return decode_QP(value)
        return value

    def __getitem__(self, header):
        return self.get(header)

    def get_all(self, header: str, default: Tuple[()] = ()) -> List[str]:
        return [decode_QP(val) for val in self.message.get_all(header, default) if val is not None]

    def is_multipart(self):
        return self.message.is_multipart()

    def get_boundary(self):
        return self.message.get_boundary()

    def walk(self):
        for part in self.message.walk():
            yield UMessage(part)

    def get_payload(
        self, index: Optional[Any] = None, decode: bool = False
    ) -> Union[str, "UMessage", List["UMessage"]]:
        message = self.message

        if index is None:
            # mypy: Argument 1 to "get_payload" of "Message" has incompatible type "None";
            # mypy: expected "int"
            # email.message.Message.get_payload has type signature:
            # Message.get_payload(self, i=None, decode=False)
            # so None seems to be totally acceptable, I don't understand mypy here
            payload = message.get_payload(index, decode)  # type: ignore

            if isinstance(payload, list):
                return [UMessage(msg) for msg in payload]

            if message.get_content_maintype() != "text":
                return payload

            if isinstance(payload, str):
                return payload

            charset = message.get_content_charset() or "iso-8859-1"
            if search_function(charset) is None:
                charset = "iso-8859-1"

            return str(payload or b"", charset, "replace")
        else:
            payload = UMessage(message.get_payload(index, decode))

        return payload

    def get_content_maintype(self):
        return str(self.message.get_content_maintype())

    def get_content_type(self):
        return str(self.message.get_content_type())

    def get_filename(self, failobj=None):
        value = self.message.get_filename(failobj)
        if value is failobj:
            return value
        try:
            return str(value)
        except UnicodeDecodeError:
            return "error decoding filename"

    # other convenience methods ###############################################

    def headers(self):
        """return an unicode string containing all the message's headers"""
        values = []
        for header in self.message.keys():
            values.append("%s: %s" % (header, self.get(header)))
        return "\n".join(values)

    def multi_addrs(self, header):
        """return a list of 2-uple (name, address) for the given address (which
        is expected to be an header containing address such as from, to, cc...)
        """
        persons = []
        for person in self.get_all(header, ()):
            name, mail = parseaddr(person)
            persons.append((name, mail))
        return persons

    def date(self, alternative_source=False, return_str=False):
        """return a datetime object for the email's date or None if no date is
        set or if it can't be parsed
        """
        value = self.get("date")
        if value is None and alternative_source:
            unix_from = self.message.get_unixfrom()
            if unix_from is not None:
                try:
                    value = unix_from.split(" ", 2)[2]
                except IndexError:
                    pass
        if value is not None:
            datetuple = parsedate(value)
            if datetuple:
                if lgc.USE_MX_DATETIME:
                    return DateTime(*datetuple[:6])
                return datetime(*datetuple[:6])
            elif not return_str:
                return None
        return value



# copyright 2003-2013 LOGILAB S.A. (Paris, FRANCE), all rights reserved.
# contact http://www.logilab.fr/ -- mailto:contact@logilab.fr
#
# This file is part of Logilab-common.
#
# Logilab-common is free software: you can redistribute it and/or modify it
# under the terms of the GNU Lesser General Public License as published by the
# Free Software Foundation, either version 2.1 of the License, or (at your
# option) any later version.
#
# Logilab-common is distributed in the hope that it will be useful, but WITHOUT
# ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
# FOR A PARTICULAR PURPOSE.  See the GNU Lesser General Public License for more
# details.
#
# You should have received a copy of the GNU Lesser General Public License along
# with Logilab-common.  If not, see <http://www.gnu.org/licenses/>.
"""This module provides bases for predicates dispatching (the pattern in use
here is similar to what's refered as multi-dispatch or predicate-dispatch in the
literature, though a bit different since the idea is to select across different
implementation 'e.g. classes), not to dispatch a message to a function or
method. It contains the following classes:

* :class:`RegistryStore`, the top level object which loads implementation
  objects and stores them into registries. You'll usually use it to access
  registries and their contained objects;

* :class:`Registry`, the base class which contains objects semantically grouped
  (for instance, sharing a same API, hence the 'implementation' name). You'll
  use it to select the proper implementation according to a context. Notice you
  may use registries on their own without using the store.

.. Note::

  implementation objects are usually designed to be accessed through the
  registry and not by direct instantiation, besides to use it as base classe.

The selection procedure is delegated to a selector, which is responsible for
scoring the object according to some context. At the end of the selection, if an
implementation has been found, an instance of this class is returned. A selector
is built from one or more predicates combined together using AND, OR, NOT
operators (actually `&`, `|` and `~`). You'll thus find some base classes to
build predicates:

* :class:`Predicate`, the abstract base predicate class

* :class:`AndPredicate`, :class:`OrPredicate`, :class:`NotPredicate`, which you
  shouldn't have to use directly. You'll use `&`, `|` and '~' operators between
  predicates directly

* :func:`objectify_predicate`

You'll eventually find one concrete predicate: :class:`yes`

.. autoclass:: RegistryStore
.. autoclass:: Registry

Predicates
----------
.. autoclass:: Predicate
.. autofunction:: objectify_predicate
.. autoclass:: yes
.. autoclass:: AndPredicate
.. autoclass:: OrPredicate
.. autoclass:: NotPredicate

Debugging
---------
.. autoclass:: traced_selection

Exceptions
----------
.. autoclass:: RegistryException
.. autoclass:: RegistryNotFound
.. autoclass:: ObjectNotFound
.. autoclass:: NoSelectableObject
"""

from __future__ import print_function


__docformat__ = "restructuredtext en"

import sys
import pkgutil
import types
import weakref
import traceback as tb
from os import listdir, stat
from os.path import join, isdir, exists
from typing import Dict, Type, Optional, Union, Sequence
from logging import getLogger
from warnings import warn
from typing import List, Tuple, Any, Iterable, Callable
from types import ModuleType

if sys.version_info >= (3, 8):
    from typing import TypedDict

    # TypedDict is also available in typing_extension > 3.7.4,
    # but this package is not available on debian buster.

from logilab.common.modutils import modpath_from_file
from logilab.common.logging_ext import set_log_methods
from logilab.common.decorators import classproperty
from logilab.common.deprecation import callable_deprecated


# selector base classes and operations ########################################


def objectify_predicate(selector_func: Callable) -> Any:
    """Most of the time, a simple score function is enough to build a selector.
    The :func:`objectify_predicate` decorator turn it into a proper selector
    class::

        @objectify_predicate
        def one(cls, req, rset=None, **kwargs):
            return 1

        class MyView(View):
            __select__ = View.__select__ & one()

    """
    return type(
        selector_func.__name__,
        (Predicate,),
        {
            "__doc__": selector_func.__doc__,
            "__call__": lambda self, *a, **kw: selector_func(*a, **kw),
        },
    )


_PREDICATES: Dict[int, Type] = {}


def wrap_predicates(decorator: Callable) -> None:
    for predicate in _PREDICATES.values():
        if "_decorators" not in predicate.__dict__:
            predicate._decorators = set()
        if decorator in predicate._decorators:
            continue
        predicate._decorators.add(decorator)
        predicate.__call__ = decorator(predicate.__call__)


class PredicateMetaClass(type):
    def __new__(mcs, *args, **kwargs):
        # use __new__ so subclasses doesn't have to call Predicate.__init__
        inst = type.__new__(mcs, *args, **kwargs)
        proxy = weakref.proxy(inst, lambda p: _PREDICATES.pop(id(p)))
        _PREDICATES[id(proxy)] = proxy
        return inst


class Predicate(object, metaclass=PredicateMetaClass):
    """base class for selector classes providing implementation
    for operators ``&``, ``|`` and  ``~``

    This class is only here to give access to binary operators, the selector
    logic itself should be implemented in the :meth:`__call__` method. Notice it
    should usually accept any arbitrary arguments (the context), though that may
    vary depending on your usage of the registry.

    a selector is called to help choosing the correct object for a
    particular context by returning a score (`int`) telling how well
    the implementation given as first argument fit to the given context.

    0 score means that the class doesn't apply.
    """

    @property
    def func_name(self):
        # backward compatibility
        return self.__class__.__name__

    def search_selector(self, selector: "Predicate") -> Optional["Predicate"]:
        """search for the given selector, selector instance or tuple of
        selectors in the selectors tree. Return None if not found.
        """
        if self is selector:
            return self
        if (isinstance(selector, type) or isinstance(selector, tuple)) and isinstance(
            self, selector
        ):
            return self
        return None

    def __str__(self):
        return self.__class__.__name__

    def __and__(self, other: "Predicate") -> "AndPredicate":
        return AndPredicate(self, other)

    def __rand__(self, other: "Predicate") -> "AndPredicate":
        return AndPredicate(other, self)

    def __iand__(self, other: "Predicate") -> "AndPredicate":
        return AndPredicate(self, other)

    def __or__(self, other: "Predicate") -> "OrPredicate":
        return OrPredicate(self, other)

    def __ror__(self, other: "Predicate"):
        return OrPredicate(other, self)

    def __ior__(self, other: "Predicate") -> "OrPredicate":
        return OrPredicate(self, other)

    def __invert__(self):
        return NotPredicate(self)

    # XXX (function | function) or (function & function) not managed yet

    def __call__(self, cls, *args, **kwargs):
        return NotImplementedError(
            "selector %s must implement its logic " "in its __call__ method" % self.__class__
        )

    def __repr__(self):
        return "<Predicate %s at %x>" % (self.__class__.__name__, id(self))


class MultiPredicate(Predicate):
    """base class for compound selector classes"""

    def __init__(self, *selectors: Any) -> None:
        self.selectors = self.merge_selectors(selectors)

    def __str__(self):
        return "%s(%s)" % (self.__class__.__name__, ",".join(str(s) for s in self.selectors))

    @classmethod
    def merge_selectors(cls, selectors: Sequence[Predicate]) -> List[Predicate]:
        """deal with selector instanciation when necessary and merge
        multi-selectors if possible:

        AndPredicate(AndPredicate(sel1, sel2), AndPredicate(sel3, sel4))
        ==> AndPredicate(sel1, sel2, sel3, sel4)
        """
        merged_selectors = []
        for selector in selectors:
            # XXX do we really want magic-transformations below?
            # if so, wanna warn about them?
            if isinstance(selector, types.FunctionType):
                selector = objectify_predicate(selector)()
            if isinstance(selector, type) and issubclass(selector, Predicate):
                selector = selector()
            assert isinstance(selector, Predicate), selector
            if isinstance(selector, cls):
                merged_selectors += selector.selectors
            else:
                merged_selectors.append(selector)
        return merged_selectors

    def search_selector(self, selector: Predicate) -> Optional[Predicate]:
        """search for the given selector or selector instance (or tuple of
        selectors) in the selectors tree. Return None if not found
        """
        for childselector in self.selectors:
            if childselector is selector:
                return childselector
            found = childselector.search_selector(selector)
            if found is not None:
                return found
        # if not found in children, maybe we are looking for self?
        return super(MultiPredicate, self).search_selector(selector)


class AndPredicate(MultiPredicate):
    """and-chained selectors"""

    def __call__(self, cls: Optional[Any], *args: Any, **kwargs: Any) -> int:
        score = 0
        for selector in self.selectors:
            partscore = selector(cls, *args, **kwargs)
            if not partscore:
                return 0
            score += partscore
        return score


class OrPredicate(MultiPredicate):
    """or-chained selectors"""

    def __call__(self, cls: Optional[Any], *args: Any, **kwargs: Any) -> int:
        for selector in self.selectors:
            partscore = selector(cls, *args, **kwargs)
            if partscore:
                return partscore
        return 0


class NotPredicate(Predicate):
    """negation selector"""

    def __init__(self, selector):
        self.selector = selector

    def __call__(self, cls, *args, **kwargs):
        score = self.selector(cls, *args, **kwargs)
        return int(not score)

    def __str__(self):
        return "NOT(%s)" % self.selector


class yes(Predicate):  # pylint: disable=C0103
    """Return the score given as parameter, with a default score of 0.5 so any
    other selector take precedence.

    Usually used for objects which can be selected whatever the context, or
    also sometimes to add arbitrary points to a score.

    Take care, `yes(0)` could be named 'no'...
    """

    def __init__(self, score: float = 0.5) -> None:
        self.score = score

    def __call__(self, *args, **kwargs):
        return self.score


# deprecated stuff #############################################################


@callable_deprecated("[lgc 0.59] use Registry.objid class method instead")
def classid(cls):
    return "%s.%s" % (cls.__module__, cls.__name__)


@callable_deprecated("[lgc 0.59] use obj_registries function instead")
def class_registries(cls, registryname):
    return obj_registries(cls, registryname)


class RegistryException(Exception):
    """Base class for registry exception."""


class RegistryNotFound(RegistryException):
    """Raised when an unknown registry is requested.

    This is usually a programming/typo error.
    """


class ObjectNotFound(RegistryException):
    """Raised when an unregistered object is requested.

    This may be a programming/typo or a misconfiguration error.
    """


class NoSelectableObject(RegistryException):
    """Raised when no object is selectable for a given context."""

    def __init__(self, args, kwargs, objects):
        self.args = args
        self.kwargs = kwargs
        self.objects = objects

    def __str__(self):
        return "args: %s, kwargs: %s\ncandidates: %s" % (
            self.args,
            self.kwargs.keys(),
            self.objects,
        )


class SelectAmbiguity(RegistryException):
    """Raised when several objects compete at selection time with an equal
    score.

    """


def _modname_from_path(path: str, extrapath: Optional[Any] = None) -> str:
    modpath = modpath_from_file(path, extrapath)
    # omit '__init__' from package's name to avoid loading that module
    # once for each name when it is imported by some other object
    # module. This supposes import in modules are done as::
    #
    #   from package import something
    #
    # not::
    #
    #   from package.__init__ import something
    #
    # which seems quite correct.
    if modpath[-1] == "__init__":
        modpath.pop()
    return ".".join(modpath)


def _toload_info(
    path: List[str], extrapath: Optional[Any], _toload: Optional[Tuple[Dict[str, str], List]] = None
) -> Tuple[Dict[str, str], List[Tuple[str, str]]]:
    """Return a dictionary of <modname>: <modpath> and an ordered list of
    (file, module name) to load
    """
    if _toload is None:
        assert isinstance(path, list)
        _toload = {}, []

    for fileordir in path:
        if isdir(fileordir) and exists(join(fileordir, "__init__.py")):
            subfiles = [join(fileordir, fname) for fname in listdir(fileordir)]

            _toload_info(subfiles, extrapath, _toload)

        elif fileordir[-3:] == ".py":
            modname = _modname_from_path(fileordir, extrapath)
            _toload[0][modname] = fileordir

            _toload[1].append((fileordir, modname))

    return _toload


class RegistrableObject(object):
    """This is the base class for registrable objects which are selected
    according to a context.

    :attr:`__registry__`
      name of the registry for this object (string like 'views',
      'templates'...). You may want to define `__registries__` directly if your
      object should be registered in several registries.

    :attr:`__regid__`
      object's identifier in the registry (string like 'main',
      'primary', 'folder_box')

    :attr:`__select__`
      class'selector

    Moreover, the `__abstract__` attribute may be set to True to indicate that a
    class is abstract and should not be registered.

    You don't have to inherit from this class to put it in a registry (having
    `__regid__` and `__select__` is enough), though this is needed for classes
    that should be automatically registered.
    """

    __registry__: Optional[str] = None
    __regid__: Optional[str] = None
    __select__: Union[None, str, Predicate] = None
    __abstract__ = True  # see doc snipppets below (in Registry class)

    @classproperty
    def __registries__(cls) -> Union[Tuple[str], Tuple]:
        if cls.__registry__ is None:
            return ()
        return (cls.__registry__,)


class RegistrableInstance(RegistrableObject):
    """Inherit this class if you want instances of the classes to be
    automatically registered.
    """

    def __new__(cls, *args, **kwargs):
        """Add a __module__ attribute telling the module where the instance was
        created, for automatic registration.
        """
        module = kwargs.pop("__module__", None)
        obj = super(RegistrableInstance, cls).__new__(cls)
        if module is None:
            warn(
                "instantiate {0} with " "__module__=__name__".format(cls.__name__),
                DeprecationWarning,
            )
            # XXX subclass must no override __new__
            filepath = tb.extract_stack(limit=2)[0][0]
            obj.__module__ = _modname_from_path(filepath)
        else:
            obj.__module__ = module
        return obj

    def __init__(self, __module__: Optional[str] = None) -> None:
        super(RegistrableInstance, self).__init__()


if sys.version_info >= (3, 8):
    SelectBestReport = TypedDict(
        "SelectBestReport",
        {
            "all_objects": List,
            "end_score": int,
            "winners": List,
            "winner": Optional[Any],
            "self": "Registry",
            "args": List,
            "kwargs": Dict,
            "registry": "Registry",
        },
    )
else:
    SelectBestReport = Dict[str, Union[List, int, Optional[Any], "Registry", Dict]]


class Registry(dict):
    """The registry store a set of implementations associated to identifier:

    * to each identifier are associated a list of implementations

    * to select an implementation of a given identifier, you should use one of the
      :meth:`select` or :meth:`select_or_none` method

    * to select a list of implementations for a context, you should use the
      :meth:`possible_objects` method

    * dictionary like access to an identifier will return the bare list of
      implementations for this identifier.

    To be usable in a registry, the only requirement is to have a `__select__`
    attribute.

    At the end of the registration process, the :meth:`__registered__`
    method is called on each registered object which have them, given the
    registry in which it's registered as argument.

    Registration methods:

    .. automethod:: register
    .. automethod:: unregister

    Selection methods:

    .. automethod:: select
    .. automethod:: select_or_none
    .. automethod:: possible_objects
    .. automethod:: object_by_id
    """

    def __init__(self, debugmode: bool) -> None:
        super(Registry, self).__init__()
        self.debugmode = debugmode
        self._select_listeners: List[Callable[[SelectBestReport], None]] = []

    def __getitem__(self, name):
        """return the registry (list of implementation objects) associated to
        this name
        """
        try:
            return super(Registry, self).__getitem__(name)
        except KeyError:
            exc = ObjectNotFound(name)
            exc.__traceback__ = sys.exc_info()[-1]
            raise exc

    @classmethod
    def objid(cls, obj: Any) -> str:
        """returns a unique identifier for an object stored in the registry"""
        return "%s.%s" % (obj.__module__, cls.objname(obj))

    @classmethod
    def objname(cls, obj: Any) -> str:
        """returns a readable name for an object stored in the registry"""
        return getattr(obj, "__name__", id(obj))

    def initialization_completed(self) -> None:
        """call method __registered__() on registered objects when the callback
        is defined"""
        for objects in self.values():
            for objectcls in objects:
                registered = getattr(objectcls, "__registered__", None)
                if registered:
                    registered(self)
        if self.debugmode:
            wrap_predicates(_lltrace)

    def register(self, obj: Any, oid: Optional[Any] = None, clear: bool = False) -> None:
        """base method to add an object in the registry"""
        assert "__abstract__" not in obj.__dict__, obj
        assert obj.__select__, obj
        oid = oid or obj.__regid__
        assert oid, (
            "no explicit name supplied to register object %s, " "which has no __regid__ set" % obj
        )
        if clear:
            objects = self[oid] = []
        else:
            objects = self.setdefault(oid, [])
        assert obj not in objects, "object %s is already registered" % obj
        objects.append(obj)

    def register_and_replace(self, obj, replaced):
        """remove <replaced> and register <obj>"""
        # XXXFIXME this is a duplication of unregister()
        # remove register_and_replace in favor of unregister + register
        # or simplify by calling unregister then register here
        if not isinstance(replaced, str):
            replaced = self.objid(replaced)
        # prevent from misspelling
        assert obj is not replaced, "replacing an object by itself: %s" % obj
        registered_objs = self.get(obj.__regid__, ())
        for index, registered in enumerate(registered_objs):
            if self.objid(registered) == replaced:
                del registered_objs[index]
                break
        else:
            self.warning("trying to replace %s that is not registered with %s", replaced, obj)
        self.register(obj)

    def unregister(self, obj):
        """remove object <obj> from this registry"""
        objid = self.objid(obj)
        oid = obj.__regid__
        for registered in self.get(oid, ()):
            # use self.objid() to compare objects because vreg will probably
            # have its own version of the object, loaded through execfile
            if self.objid(registered) == objid:
                self[oid].remove(registered)
                break
        else:
            self.warning("can't remove %s, no id %s in the registry", objid, oid)

    def all_objects(self):
        """return a list containing all objects in this registry."""
        result = []
        for objs in self.values():
            result += objs
        return result

    # dynamic selection methods ################################################

    def object_by_id(self, oid, *args, **kwargs):
        """return object with the `oid` identifier. Only one object is expected
        to be found.

        raise :exc:`ObjectNotFound` if there are no object with id `oid` in this
        registry

        raise :exc:`AssertionError` if there is more than one object there
        """
        objects = self[oid]
        assert len(objects) == 1, objects
        return objects[0](*args, **kwargs)

    def select(self, __oid, *args, **kwargs):
        """return the most specific object among those with the given oid
        according to the given context.

        raise :exc:`ObjectNotFound` if there are no object with id `oid` in this
        registry

        raise :exc:`NoSelectableObject` if no object can be selected
        """
        obj = self._select_best(self[__oid], *args, **kwargs)
        if obj is None:
            raise NoSelectableObject(args, kwargs, self[__oid])
        return obj

    def select_or_none(self, __oid, *args, **kwargs):
        """return the most specific object among those with the given oid
        according to the given context, or None if no object applies.
        """
        try:
            return self._select_best(self[__oid], *args, **kwargs)
        except ObjectNotFound:
            return None

    def possible_objects(self, *args, **kwargs):
        """return an iterator on possible objects in this registry for the given
        context
        """
        for objects in self.values():
            obj = self._select_best(objects, *args, **kwargs)
            if obj is None:
                continue
            yield obj

    def add_select_best_listener(self, listener):
        """
        Add a listener to the list of one parameters callables (function/method)
        that will be called everytime the selection Registry._select_best is
        done and they will recieve a dict of the following form::

          {"all_objects": [], "end_score": 0, "winners": [], "winner": None or winner,
          "self": self, "args": args, "kwargs": kwargs, }
        """
        self._select_listeners.append(listener)

    def _select_best(self, objects, *args, **kwargs):
        """return an instance of the most specific object according
        to parameters

        return None if not object apply (don't raise `NoSelectableObject` since
        it's costly when searching objects using `possible_objects`
        (e.g. searching for hooks).

        Every callable stored in Registry._select_listeners will be called
        once the selection is done and will recieve a dict of the following form::

          {"all_objects": [], "end_score": 0, "winners": [], "winner": None or winner,
          "self": self, "args": args, "kwargs": kwargs, "registry": self}
        """

        if self._select_listeners:
            select_best_report: SelectBestReport = {
                "registry": self,
                "all_objects": [],
                "end_score": 0,
                "winners": [],
                "winner": None,
                "self": self,
                "args": args,
                "kwargs": kwargs,
            }

        score, winners = 0, None
        for obj in objects:
            objectscore = obj.__select__(obj, *args, **kwargs)

            if self._select_listeners:
                select_best_report["all_objects"].append({"object": obj, "score": objectscore})

            if objectscore > score:
                score, winners = objectscore, [obj]

            elif objectscore > 0 and objectscore == score:
                winners.append(obj)

        if self._select_listeners:
            select_best_report["winners"] = winners.copy() if winners else []
            select_best_report["winner"] = winners[0] if winners else winners
            select_best_report["end_score"] = score
            for listener in self._select_listeners:
                listener(select_best_report)

        if winners is None:
            return None

        if len(winners) > 1:
            # log in production environement / test, error while debugging
            msg = "select ambiguity: %s\n(args: %s, kwargs: %s)"

            if self.debugmode:
                # raise bare exception in debug mode
                raise SelectAmbiguity(msg % (winners, args, kwargs.keys()))

            self.error(msg, winners, args, kwargs.keys())

        # return the result of calling the object
        return self.selected(winners[0], args, kwargs)

    def selected(self, winner, args, kwargs):
        """override here if for instance you don't want "instanciation" """
        return winner(*args, **kwargs)

    # these are overridden by set_log_methods below
    # only defining here to prevent pylint from complaining
    info = warning = error = critical = exception = debug = lambda msg, *a, **kw: None


def obj_registries(cls: Any, registryname: Optional[Any] = None) -> Tuple[str]:
    """return a tuple of registry names (see __registries__)"""
    if registryname:
        return (registryname,)
    return cls.__registries__


class RegistryStore(dict):
    """This class is responsible for loading objects and storing them
    in their registry which is created on the fly as needed.

    It handles dynamic registration of objects and provides a
    convenient api to access them. To be recognized as an object that
    should be stored into one of the store's registry
    (:class:`Registry`), an object must provide the following
    attributes, used control how they interact with the registry:

    :attr:`__registries__`
      list of registry names (string like 'views', 'templates'...) into which
      the object should be registered

    :attr:`__regid__`
      object identifier in the registry (string like 'main',
      'primary', 'folder_box')

    :attr:`__select__`
      the object predicate selectors

    Moreover, the :attr:`__abstract__` attribute may be set to `True`
    to indicate that an object is abstract and should not be registered
    (such inherited attributes not considered).

    .. Note::

      When using the store to load objects dynamically, you *always* have
      to use **super()** to get the methods and attributes of the
      superclasses, and not use the class identifier. If not, you'll get into
      trouble at reload time.

      For example, instead of writing::

          class Thing(Parent):
              __regid__ = 'athing'
              __select__ = yes()

              def f(self, arg1):
                  Parent.f(self, arg1)

      You must write::

          class Thing(Parent):
              __regid__ = 'athing'
              __select__ = yes()

              def f(self, arg1):
                  super(Thing, self).f(arg1)

    Controlling object registration
    -------------------------------

    Dynamic loading is triggered by calling the :meth:`register_modnames`
    method, given a list of modules names to inspect.

    .. automethod:: register_modnames

    For each module, by default, all compatible objects are registered
    automatically. However if some objects come as replacement of
    other objects, or have to be included only if some condition is
    met, you'll have to define a `registration_callback(vreg)`
    function in the module and explicitly register **all objects** in
    this module, using the api defined below.


    .. automethod:: RegistryStore.register_all
    .. automethod:: RegistryStore.register_and_replace
    .. automethod:: RegistryStore.register
    .. automethod:: RegistryStore.unregister

    .. Note::
        Once the function `registration_callback(vreg)` is implemented in a
        module, all the objects from this module have to be explicitly
        registered as it disables the automatic object registration.


    Examples:

    .. sourcecode:: python

       def registration_callback(store):
          # register everything in the module except BabarClass
          store.register_all(globals().values(), __name__, (BabarClass,))

          # conditionally register BabarClass
          if 'babar_relation' in store.schema:
              store.register(BabarClass)

    In this example, we register all application object classes defined in the module
    except `BabarClass`. This class is then registered only if the 'babar_relation'
    relation type is defined in the instance schema.

    .. sourcecode:: python

       def registration_callback(store):
          store.register(Elephant)
          # replace Babar by Celeste
          store.register_and_replace(Celeste, Babar)

    In this example, we explicitly register classes one by one:

    * the `Elephant` class
    * the `Celeste` to replace `Babar`

    If at some point we register a new appobject class in this module, it won't be
    registered at all without modification to the `registration_callback`
    implementation. The first example will register it though, thanks to the call
    to the `register_all` method.

    Controlling registry instantiation
    ----------------------------------

    The `REGISTRY_FACTORY` class dictionary allows to specify which class should
    be instantiated for a given registry name. The class associated to `None`
    key will be the class used when there is no specific class for a name.
    """

    def __init__(self, debugmode: bool = False) -> None:
        super(RegistryStore, self).__init__()
        self.debugmode = debugmode

    def reset(self) -> None:
        """clear all registries managed by this store"""
        # don't use self.clear, we want to keep existing subdictionaries
        for subdict in self.values():
            subdict.clear()
        self._lastmodifs: Dict[str, int] = {}

    def __getitem__(self, name: str) -> Registry:
        """return the registry (dictionary of class objects) associated to
        this name
        """
        try:
            return super(RegistryStore, self).__getitem__(name)
        except KeyError:
            exc = RegistryNotFound(name)
            exc.__traceback__ = sys.exc_info()[-1]
            raise exc

    # methods for explicit (un)registration ###################################

    # default class, when no specific class set
    REGISTRY_FACTORY: Dict[Union[None, str], type] = {None: Registry}

    def registry_class(self, regid: str) -> type:
        """return existing registry named regid or use factory to create one and
        return it"""
        try:
            return self.REGISTRY_FACTORY[regid]
        except KeyError:
            return self.REGISTRY_FACTORY[None]

    # mypy: Signature of "setdefault" incompatible with supertype "MutableMapping""
    # I don't know how to overload signatures of method in mypy
    def setdefault(self, regid: str) -> Registry:  # type: ignore
        try:
            return self[regid]
        except RegistryNotFound:
            self[regid] = self.registry_class(regid)(self.debugmode)
            return self[regid]

    def register_all(self, objects: Iterable, modname: str, butclasses: Sequence = ()) -> None:
        """register registrable objects into `objects`.

        Registrable objects are properly configured subclasses of
        :class:`RegistrableObject`.  Objects which are not defined in the module
        `modname` or which are in `butclasses` won't be registered.

        Typical usage is:

        .. sourcecode:: python

            store.register_all(globals().values(), __name__, (ClassIWantToRegisterExplicitly,))

        So you get partially automatic registration, keeping manual registration
        for some object (to use
        :meth:`~logilab.common.registry.RegistryStore.register_and_replace` for
        instance).
        """
        assert isinstance(modname, str), (
            "modname expected to be a module name (ie string), got %r" % modname
        )
        for obj in objects:
            if self.is_registrable(obj) and obj.__module__ == modname and obj not in butclasses:
                if isinstance(obj, type):
                    self._load_ancestors_then_object(modname, obj, butclasses)
                else:
                    self.register(obj)

    def register(
        self,
        obj: Any,
        registryname: Optional[Any] = None,
        oid: Optional[Any] = None,
        clear: bool = False,
    ) -> None:
        """register `obj` implementation into `registryname` or
        `obj.__registries__` if not specified, with identifier `oid` or
        `obj.__regid__` if not specified.

        If `clear` is true, all objects with the same identifier will be
        previously unregistered.
        """
        assert not obj.__dict__.get("__abstract__"), obj
        for registryname in obj_registries(obj, registryname):
            registry = self.setdefault(registryname)
            registry.register(obj, oid=oid, clear=clear)
            self.debug(
                "register %s in %s['%s']", registry.objname(obj), registryname, oid or obj.__regid__
            )
            self._loadedmods.setdefault(obj.__module__, {})[registry.objid(obj)] = obj

    def unregister(self, obj, registryname=None):
        """unregister `obj` object from the registry `registryname` or
        `obj.__registries__` if not specified.
        """
        for registryname in obj_registries(obj, registryname):
            registry = self[registryname]
            registry.unregister(obj)
            self.debug(
                "unregister %s from %s['%s']", registry.objname(obj), registryname, obj.__regid__
            )

    def register_and_replace(self, obj, replaced, registryname=None):
        """register `obj` object into `registryname` or
        `obj.__registries__` if not specified. If found, the `replaced` object
        will be unregistered first (else a warning will be issued as it is
        generally unexpected).
        """
        for registryname in obj_registries(obj, registryname):
            registry = self[registryname]
            registry.register_and_replace(obj, replaced)
            self.debug(
                "register %s in %s['%s'] instead of %s",
                registry.objname(obj),
                registryname,
                obj.__regid__,
                registry.objname(replaced),
            )

    # initialization methods ###################################################

    def init_registration(
        self, path: List[str], extrapath: Optional[Any] = None
    ) -> List[Tuple[str, str]]:
        """reset registry and walk down path to return list of (path, name)
        file modules to be loaded"""
        # XXX make this private by renaming it to _init_registration ?
        self.reset()
        # compute list of all modules that have to be loaded
        self._toloadmods, filemods = _toload_info(path, extrapath)
        # XXX is _loadedmods still necessary ? It seems like it's useful
        #     to avoid loading same module twice, especially with the
        #     _load_ancestors_then_object logic but this needs to be checked
        self._loadedmods: Dict[str, Dict[str, type]] = {}
        return filemods

    @callable_deprecated("use register_modnames() instead")
    def register_objects(self, path: List[str], extrapath: Optional[Any] = None) -> None:
        """register all objects found walking down <path>"""
        # load views from each directory in the instance's path
        # XXX inline init_registration ?
        filemods = self.init_registration(path, extrapath)
        for filepath, modname in filemods:
            self.load_file(filepath, modname)
        self.initialization_completed()

    def register_modnames(self, modnames: List[str]) -> None:
        """register all objects found in <modnames>"""
        self.reset()
        self._loadedmods = {}
        self._toloadmods = {}
        toload = []
        for modname in modnames:
            loader = pkgutil.find_loader(modname)
            assert loader is not None
            # mypy: "Loader" has no attribute "get_filename"
            # the selected class has one
            filepath = loader.get_filename()  # type: ignore
            if filepath[-4:] in (".pyc", ".pyo"):
                # The source file *must* exists
                filepath = filepath[:-1]
            self._toloadmods[modname] = filepath
            toload.append((filepath, modname))
        for filepath, modname in toload:
            self.load_file(filepath, modname)
        self.initialization_completed()

    def initialization_completed(self) -> None:
        """call initialization_completed() on all known registries"""
        for reg in self.values():
            reg.initialization_completed()

    def _mdate(self, filepath: str) -> Optional[int]:
        """ return the modification date of a file path """
        try:
            return stat(filepath)[-2]
        except OSError:
            # this typically happens on emacs backup files (.#foo.py)
            self.warning("Unable to load %s. It is likely to be a backup file", filepath)
            return None

    def is_reload_needed(self, path):
        """return True if something module changed and the registry should be
        reloaded
        """
        lastmodifs = self._lastmodifs
        for fileordir in path:
            if isdir(fileordir) and exists(join(fileordir, "__init__.py")):
                if self.is_reload_needed([join(fileordir, fname) for fname in listdir(fileordir)]):
                    return True
            elif fileordir[-3:] == ".py":
                mdate = self._mdate(fileordir)
                if mdate is None:
                    continue  # backup file, see _mdate implementation
                elif "flymake" in fileordir:
                    # flymake + pylint in use, don't consider these they will corrupt the registry
                    continue
                if fileordir not in lastmodifs or lastmodifs[fileordir] < mdate:
                    self.info("File %s changed since last visit", fileordir)
                    return True
        return False

    def load_file(self, filepath: str, modname: str) -> None:
        """ load registrable objects (if any) from a python file """
        if modname in self._loadedmods:
            return
        self._loadedmods[modname] = {}
        mdate = self._mdate(filepath)
        if mdate is None:
            return  # backup file, see _mdate implementation
        elif "flymake" in filepath:
            # flymake + pylint in use, don't consider these they will corrupt the registry
            return
        # set update time before module loading, else we get some reloading
        # weirdness in case of syntax error or other error while importing the
        # module
        self._lastmodifs[filepath] = mdate
        # load the module
        if sys.version_info < (3,) and not isinstance(modname, str):
            modname = str(modname)
        module = __import__(modname, fromlist=modname.split(".")[:-1])
        self.load_module(module)

    def load_module(self, module: ModuleType) -> None:
        """Automatically handle module objects registration.

        Instances are registered as soon as they are hashable and have the
        following attributes:

        * __regid__ (a string)
        * __select__ (a callable)
        * __registries__ (a tuple/list of string)

        For classes this is a bit more complicated :

        - first ensure parent classes are already registered

        - class with __abstract__ == True in their local dictionary are skipped

        - object class needs to have registries and identifier properly set to a
          non empty string to be registered.
        """
        self.info("loading %s from %s", module.__name__, module.__file__)
        if hasattr(module, "registration_callback"):
            # mypy: Module has no attribute "registration_callback"
            # we check that before
            module.registration_callback(self)  # type: ignore
        else:
            self.register_all(vars(module).values(), module.__name__)

    def _load_ancestors_then_object(
        self, modname: str, objectcls: type, butclasses: Sequence[Any] = ()
    ) -> None:
        """handle class registration according to rules defined in
        :meth:`load_module`
        """
        # backward compat, we used to allow whatever else than classes
        if not isinstance(objectcls, type):
            if self.is_registrable(objectcls) and objectcls.__module__ == modname:
                self.register(objectcls)
            return
        # imported classes
        objmodname = objectcls.__module__
        if objmodname != modname:
            # The module of the object is not the same as the currently
            # worked on module, or this is actually an instance, which
            # has no module at all
            if objmodname in self._toloadmods:
                # if this is still scheduled for loading, let's proceed immediately,
                # but using the object module
                self.load_file(self._toloadmods[objmodname], objmodname)
            return
        # ensure object hasn't been already processed
        clsid = "%s.%s" % (modname, objectcls.__name__)
        if clsid in self._loadedmods[modname]:
            return
        self._loadedmods[modname][clsid] = objectcls
        # ensure ancestors are registered
        for parent in objectcls.__bases__:
            self._load_ancestors_then_object(modname, parent, butclasses)
        # ensure object is registrable
        if objectcls in butclasses or not self.is_registrable(objectcls):
            return
        # backward compat
        reg = self.setdefault(obj_registries(objectcls)[0])
        if reg.objname(objectcls)[0] == "_":
            warn(
                "[lgc 0.59] object whose name start with '_' won't be "
                "skipped anymore at some point, use __abstract__ = True "
                "instead (%s)" % objectcls,
                DeprecationWarning,
            )
            return
        # register, finally
        self.register(objectcls)

    @classmethod
    def is_registrable(cls, obj: Any) -> bool:
        """ensure `obj` should be registered

        as arbitrary stuff may be registered, do a lot of check and warn about
        weird cases (think to dumb proxy objects)
        """
        if isinstance(obj, type):
            if not issubclass(obj, RegistrableObject):
                # ducktyping backward compat
                if not (
                    getattr(obj, "__registries__", None)
                    and getattr(obj, "__regid__", None)
                    and getattr(obj, "__select__", None)
                ):
                    return False
            elif issubclass(obj, RegistrableInstance):
                return False

        elif not isinstance(obj, RegistrableInstance):
            return False

        if not obj.__regid__:
            return False  # no regid

        registries = obj.__registries__
        if not registries:
            return False  # no registries

        selector = obj.__select__
        if not selector:
            return False  # no selector

        if obj.__dict__.get("__abstract__", False):
            return False

        # then detect potential problems that should be warned
        if not isinstance(registries, (tuple, list)):
            cls.warning("%s has __registries__ which is not a list or tuple", obj)
            return False

        if not callable(selector):
            cls.warning("%s has not callable __select__", obj)
            return False

        return True

    # these are overridden by set_log_methods below
    # only defining here to prevent pylint from complaining
    info = warning = error = critical = exception = debug = lambda msg, *a, **kw: None


# init logging
set_log_methods(RegistryStore, getLogger("registry.store"))
set_log_methods(Registry, getLogger("registry"))


# helpers for debugging selectors
TRACED_OIDS = None


def _trace_selector(cls, selector, args, ret):
    vobj = args[0]
    if TRACED_OIDS == "all" or vobj.__regid__ in TRACED_OIDS:
        print("%s -> %s for %s(%s)" % (cls, ret, vobj, vobj.__regid__))


def _lltrace(selector):
    """use this decorator on your predicates so they become traceable with
    :class:`traced_selection`
    """

    def traced(cls, *args, **kwargs):
        ret = selector(cls, *args, **kwargs)
        if TRACED_OIDS is not None:
            _trace_selector(cls, selector, args, ret)
        return ret

    traced.__name__ = selector.__name__
    traced.__doc__ = selector.__doc__
    return traced


class traced_selection(object):  # pylint: disable=C0103
    """
    Typical usage is :

    .. sourcecode:: python

        >>> from logilab.common.registry import traced_selection
        >>> with traced_selection():
        ...     # some code in which you want to debug selectors
        ...     # for all objects

    This will yield lines like this in the logs::

        selector one_line_rset returned 0 for <class 'elephant.Babar'>

    You can also give to :class:`traced_selection` the identifiers of objects on
    which you want to debug selection ('oid1' and 'oid2' in the example above).

    .. sourcecode:: python

        >>> with traced_selection( ('regid1', 'regid2') ):
        ...     # some code in which you want to debug selectors
        ...     # for objects with __regid__ 'regid1' and 'regid2'

    A potentially useful point to set up such a tracing function is
    the `logilab.common.registry.Registry.select` method body.
    """

    def __init__(self, traced="all"):
        self.traced = traced

    def __enter__(self):
        global TRACED_OIDS
        TRACED_OIDS = self.traced

    def __exit__(self, exctype, exc, traceback):
        global TRACED_OIDS
        TRACED_OIDS = None
        return traceback is None



# copyright 2003-2011 LOGILAB S.A. (Paris, FRANCE), all rights reserved.
# contact http://www.logilab.fr/ -- mailto:contact@logilab.fr
#
# This file is part of logilab-common.
#
# logilab-common is free software: you can redistribute it and/or modify it
# under the terms of the GNU Lesser General Public License as published by the
# Free Software Foundation, either version 2.1 of the License,
# or (at your option) any later version.
#
# logilab-common is distributed in the hope that it will be useful, but WITHOUT
# ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
# FOR A PARTICULAR PURPOSE.  See the GNU Lesser General Public License for more
# details.
#
# You should have received a copy of the GNU Lesser General Public License
# along with logilab-common.  If not, see <http://www.gnu.org/licenses/>.
"""HTML formatting drivers for ureports"""
__docformat__ = "restructuredtext en"


from logilab.common.ureports import BaseWriter
from logilab.common.ureports.nodes import (
    Section,
    Title,
    Table,
    List,
    Paragraph,
    Link,
    VerbatimText,
    Text,
)
from typing import Any


class HTMLWriter(BaseWriter):
    """format layouts as HTML"""

    def __init__(self, snippet: int = None) -> None:
        super(HTMLWriter, self).__init__()
        self.snippet = snippet

    def handle_attrs(self, layout: Any) -> str:
        """get an attribute string from layout member attributes"""
        attrs = ""
        klass = getattr(layout, "klass", None)
        if klass:
            attrs += ' class="%s"' % klass
        nid = getattr(layout, "id", None)
        if nid:
            attrs += ' id="%s"' % nid
        return attrs

    def begin_format(self, layout: Any) -> None:
        """begin to format a layout"""
        super(HTMLWriter, self).begin_format(layout)
        if self.snippet is None:
            self.writeln("<html>")
            self.writeln("<body>")

    def end_format(self, layout: Any) -> None:
        """finished to format a layout"""
        if self.snippet is None:
            self.writeln("</body>")
            self.writeln("</html>")

    def visit_section(self, layout: Section) -> None:
        """display a section as html, using div + h[section level]"""
        self.section += 1
        self.writeln("<div%s>" % self.handle_attrs(layout))
        self.format_children(layout)
        self.writeln("</div>")
        self.section -= 1

    def visit_title(self, layout: Title) -> None:
        """display a title using <hX>"""
        self.write("<h%s%s>" % (self.section, self.handle_attrs(layout)))
        self.format_children(layout)
        self.writeln("</h%s>" % self.section)

    def visit_table(self, layout: Table) -> None:
        """display a table as html"""
        self.writeln("<table%s>" % self.handle_attrs(layout))
        table_content = self.get_table_content(layout)
        for i in range(len(table_content)):
            row = table_content[i]
            if i == 0 and layout.rheaders:
                self.writeln('<tr class="header">')
            elif i + 1 == len(table_content) and layout.rrheaders:
                self.writeln('<tr class="header">')
            else:
                self.writeln('<tr class="%s">' % (i % 2 and "even" or "odd"))
            for j in range(len(row)):
                cell = row[j] or "&#160;"
                if (
                    (layout.rheaders and i == 0)
                    or (layout.cheaders and j == 0)
                    or (layout.rrheaders and i + 1 == len(table_content))
                    or (layout.rcheaders and j + 1 == len(row))
                ):
                    self.writeln("<th>%s</th>" % cell)
                else:
                    self.writeln("<td>%s</td>" % cell)
            self.writeln("</tr>")
        self.writeln("</table>")

    def visit_list(self, layout: List) -> None:
        """display a list as html"""
        self.writeln("<ul%s>" % self.handle_attrs(layout))
        for row in list(self.compute_content(layout)):
            self.writeln("<li>%s</li>" % row)
        self.writeln("</ul>")

    def visit_paragraph(self, layout: Paragraph) -> None:
        """display links (using <p>)"""
        self.write("<p>")
        self.format_children(layout)
        self.write("</p>")

    def visit_span(self, layout):
        """display links (using <p>)"""
        self.write("<span%s>" % self.handle_attrs(layout))
        self.format_children(layout)
        self.write("</span>")

    def visit_link(self, layout: Link) -> None:
        """display links (using <a>)"""
        self.write(' <a href="%s"%s>%s</a>' % (layout.url, self.handle_attrs(layout), layout.label))

    def visit_verbatimtext(self, layout: VerbatimText) -> None:
        """display verbatim text (using <pre>)"""
        self.write("<pre>")
        self.write(layout.data.replace("&", "&amp;").replace("<", "&lt;"))
        self.write("</pre>")

    def visit_text(self, layout: Text) -> None:
        """add some text"""
        data = layout.data
        if layout.escaped:
            data = data.replace("&", "&amp;").replace("<", "&lt;")
        self.write(data)
